// LiveHatching A^C`GtFNgvO@openCV
// program T.Ogura
// Data 2008.12.16 EMail:t-ogura@t3.rim.or.jp
//
// LvOAviUtlvOC

#include <windows.h>
#include "cv.h"
#include "highgui.h"
#include <stdio.h>		// printf() / scanf()
#define _USE_MATH_DEFINES
#include <math.h>
#include "filter.h"

#include <string>
#include <map>

#ifndef _DEBUG
#pragma comment(lib,"cv.lib")
#pragma comment(lib,"cvaux.lib")
#pragma comment(lib,"cxcore.lib")
#pragma comment(lib,"highgui.lib")
#else
#pragma comment(lib,"cvd.lib")
#pragma comment(lib,"cvauxd.lib")
#pragma comment(lib,"cxcored.lib")
#pragma comment(lib,"highguid.lib")
#endif

#define YC2R(Y,Cb,Cr) (Y + 1.402*Cr)
#define YC2G(Y,Cb,Cr) (Y - 0.344*Cb - 0.714*Cr)
#define YC2B(Y,Cb,Cr) (Y + 1.772*Cb)
#define RGB2Y(R,G,B) (0.299*R + 0.587*G + 0.114*B)
#define RGB2CB(R,G,B) (-0.169*R - 0.331*G + 0.500*B)
#define RGB2CR(R,G,B) (0.500*R - 0.419*G - 0.081*B)

// [min, max] ͈̔͂ɖOa
template <class T>
T Saturate(T v, T min, T max) {
	if (v < min) {
		return min;
	} else if (v > max) {
		return max;
	}
	return v;
}

class BufToCv {
public:
	static const double MLT;
	static const double DIV;
	
	template <class T>
	static T to(double v) {
		return (T)v;
	}

	template <>
	static unsigned char to<unsigned char>(double v) {
		return (unsigned char)Saturate(v/16.0, 0.0, 255.0);
	}

	template <>
	static float to<float>(double v) {
		return (float)(v * DIV);
	}

	template <>
	static double to<double>(double v) {
		return v * DIV;
	}

	template <class T>
	static double from(T v) {
		return v;
	}

	template <>
	static double from<unsigned char>(unsigned char v) {
		return v * 16;
	}

	template <>
	static double from<float>(float v) {
		return v * MLT;
	}

	template <>
	static double from<double>(double v) {
		return v * MLT;
	}
};

const double BufToCv::MLT = 4096.0;
const double BufToCv::DIV = 1 / BufToCv::MLT;

template <class T>
void CopyBufToCvBGRImg(
		PIXEL_YC *buf, IplImage *img,
		int x0, int y0, int w, int h, int x1, int y1, int mw, int mh
){
	int src,ix0,iy0,ix1,iy1;

	for(iy0 = y0; iy0 < y0 + h; iy0++) {
		iy1 = y1 + (iy0-y0);
		src = x0 + iy0 * mw;
		for(ix0 = x0,ix1 = x1; ix0 < x0 + w; ix0++) {
			CV_IMAGE_ELEM(img, T, iy1, ix1*3+0) = BufToCv::to<T>(YC2B(buf[src].y, buf[src].cb, buf[src].cr));
			CV_IMAGE_ELEM(img, T, iy1, ix1*3+1) = BufToCv::to<T>(YC2G(buf[src].y, buf[src].cb, buf[src].cr));
			CV_IMAGE_ELEM(img, T, iy1, ix1*3+2) = BufToCv::to<T>(YC2R(buf[src].y, buf[src].cb, buf[src].cr));
			src++;
			ix1++;
		}
	}
}

template <class T>
void CopyCvBGRImgToBuf(
		IplImage *img, PIXEL_YC *buf, 
		int x0, int y0, int w, int h, int x1, int y1, int mw, int mh
){
	int dst,ix0,iy0,ix1,iy1;
	PIXEL_YC yc;

	for(iy0 = y0; iy0 < y0 + h; iy0++) {
		iy1 = y1 + (iy0-y0);
		dst = x1 + iy1 * mw;
		for(ix0 = x0,ix1 = x1; ix0 < x0 + w; ix0++) {
			double r,g,b;
			b = BufToCv::from<T>(CV_IMAGE_ELEM(img, T, iy0, ix0*3+0));
			g = BufToCv::from<T>(CV_IMAGE_ELEM(img, T, iy0, ix0*3+1));
			r = BufToCv::from<T>(CV_IMAGE_ELEM(img, T, iy0, ix0*3+2));
			yc.y =  (short)RGB2Y( r, g, b);
			yc.cb = (short)RGB2CB(r, g, b);
			yc.cr = (short)RGB2CR(r, g, b);
			buf[dst] = yc;
			dst++;
			ix1++;
		}
	}
}

inline int is_neg(int v) {
    return (((unsigned int)v) >> (sizeof(v)*8 - 1));
}

inline int modulo(int a, int b) {
	int m = a % b;
	return m + b * is_neg(m);
}

namespace bicubic {
	static const double a = -1.0;

	inline double cubic1(double x) {
		return (((a+2.0)*x - (a+3.0))*x)*x + 1.0;
	}
	inline double cubic2(double x) {
		return ((a * x - 5.0*a) * x + 8.0*a) * x - 4.0*a;
	}

	inline double cubic_sub(unsigned char *img,
							int x0, int x1, int x2, int x3,
							double w0, double w1, double w2, double w3) {
		double v0 = (double)img[x0];
		double v1 = (double)img[x1];
		double v2 = (double)img[x2];
		double v3 = (double)img[x3];
		return v0 * w0 + v1 * w1 + v2 * w2 + v3 * w3;
	}

	inline double v_bicubic(IplImage *img, double x, double y) {
		int zx = (int)x;
		int zy = (int)y;
		double rx = x - (double)zx;
		double ry = y - (double)zy;

		if (rx < 0.0) {
			zx--;
			rx += 1.0;
		}
		if (ry < 0.0) {
			zy--;
			ry += 1.0;
		}

		double wx0 = cubic2(1.0 + rx);
		double wx1 = cubic1(      rx);
		double wx2 = cubic1(1.0 - rx);
		double wx3 = cubic2(2.0 - rx);

		double wy0 = cubic2(1.0 + ry);
		double wy1 = cubic1(      ry);
		double wy2 = cubic1(1.0 - ry);
		double wy3 = cubic2(2.0 - ry);

		int zx0 = modulo(zx-1, img->width);
		int zx1 = (zx0 + 1) % img->width;
		int zx2 = (zx0 + 2) % img->width;
		int zx3 = (zx0 + 3) % img->width;

		int zy0 = modulo(zy-1, img->height);
		int zy1 = (zy0 + 1) % img->height;
		int zy2 = (zy0 + 2) % img->height;
		int zy3 = (zy0 + 3) % img->height;

		unsigned char *dat = (unsigned char*)img->imageData;
		int widthStep = img->widthStep;

		double vy0 = cubic_sub(dat + zy0 * widthStep, zx0,zx1,zx2,zx3, wx0,wx1,wx2,wx3);
		double vy1 = cubic_sub(dat + zy1 * widthStep, zx0,zx1,zx2,zx3, wx0,wx1,wx2,wx3);
		double vy2 = cubic_sub(dat + zy2 * widthStep, zx0,zx1,zx2,zx3, wx0,wx1,wx2,wx3);
		double vy3 = cubic_sub(dat + zy3 * widthStep, zx0,zx1,zx2,zx3, wx0,wx1,wx2,wx3);

		return vy0*wy0 + vy1*wy1 + vy2*wy2 + vy3*wy3;
	}
}
using namespace bicubic;


inline double v_bilinear(IplImage *img, double x, double y) {
	int zx = (int)x;
	int zy = (int)y;
	double rx = x - (double)zx;
	double ry = y - (double)zy;

	if (rx < 0.0) {
		zx--;
		rx += 1.0;
	}
	if (ry < 0.0) {
		zy--;
		ry += 1.0;
	}

	int zx0 = modulo(zx, img->width);
	int zy0 = modulo(zy, img->height);

	int zx1 = (zx0 + 1) % img->width;
	int zy1 = (zy0 + 1) % img->height;

	int widthStep = img->widthStep;

	double v00 = (double)(unsigned char)img->imageData[zy0 * widthStep + zx0];
	double v01 = (double)(unsigned char)img->imageData[zy0 * widthStep + zx1];
	double v10 = (double)(unsigned char)img->imageData[zy1 * widthStep + zx0];
	double v11 = (double)(unsigned char)img->imageData[zy1 * widthStep + zx1];

	return ((v00 * (1.0-rx) + v01 * rx) * (1.0-ry)) + ((v10 * (1.0-rx) + v11 * rx) * ry);
}


namespace param{
	int X = 5;			// LineStep 傫ƃKNKN
	int Y = 54;			// oo@傫Ɠ͑a
	int Z = 0;			// LineLength	
	int lineWidth = 1;	// ̑
	int lineNoise = 15;	// ߐڃsNZƂ̔r臒lB0ȏɂƐr؂
	double DARK = 1.0;	// GbW̐F̋Px

	int VFlag = 0;		// K}␳݂ȁH
	int hatchingType = 0;     // 0..normal 1..inverse 2..white-base
	const int nPattern = 7;   // nb`Op^[
	int drawMode = 3;
	int drawFps = 0;
	int patternOffset = 0;
	IplImage *hatching_org[nPattern];
	IplImage *hatching[nPattern];

	double inclination = -1.0;	//X
	double incRange = 1.0;		//Xϓ
	int	xRange  = 10000;	//xϓ
	int	yRange  = 10000;	//yϓ
	double hDenseMax  = 0;		//nb`OZx

	double mDARK = 1.0;	// GbW̐F̋PxW
	double addDARK = 0.0;

	bool hatchingIntrp = false; //nb`Op^[QƎ̕ԗL

	enum INTR_TYPE{
		INTR_NONE,
		INTR_BILINEAR,
		INTR_BICUBIC
	};
	enum INTR_TYPE hatchingIntrType = INTR_NONE; //nb`Op^[QƎ̕ԃ^Cv

	int hatchingPtNo = -100;	//nb`Op^[̎

	double hatchingMagn = 1.0;  //nb`Op^[̊g嗦
	double hatchingRad  = 1.0;  //nb`Op^[̉]px(WA)

	struct hatchingTransform { //nb`Op^[ϊpW
		double a00;
		double a01;
		double a02;
		double a10;
		double a11;
		double a12;
	};
	hatchingTransform hatchingTf;
}
using namespace param;

static IplImage *img0 = NULL;
static IplImage *img1 = NULL;
static IplImage *hsv = NULL;  
static IplImage *grayImage = NULL;
static IplImage *binaryImage = NULL;

typedef struct PointTable{
	CvPoint p;
	CvScalar color;
} POINTTABLE;

static const int MAXPOINT = 100000;
static POINTTABLE pointTable[MAXPOINT];

// RGB̍v
double colorLen(CvScalar &x, CvScalar &y)
{
	double sum = 0;
	for(int i=0;i<3;i++){
		sum += abs(x.val[i] - y.val[i]);
	}
	return sum;
}

// image̍WʒuJ[
CvScalar getColor(IplImage* image, int x, int y)
{
	int index = y * image->widthStep + x * image->nChannels;
	unsigned char *data =  (unsigned char *)image->imageData;
	return CV_RGB(data[index+2],data[index+1],data[index+0]);
}

// WC[WOɂ\̂ꍇp
CvScalar getColorCheck(IplImage* image, int x, int y)
{
	if (x<0) x = 0;
	if (y<0) y = 0;
	if (x>=image->width) x = image->width-1;
	if (y>=image->height) y = image->height-1;
	return getColor(image, x, y);
}

void divScaler(CvScalar &s, double d)
{
	s.val[0] /= d;
	s.val[1] /= d;
	s.val[2] /= d;
	s.val[3] /= d;
}
void mulScaler(CvScalar &s, double d)
{
	s.val[0] *= d;
	s.val[1] *= d;
	s.val[2] *= d;
	s.val[3] *= d;
}

void muladdScaler(CvScalar &s, double d, double a)
{
	s.val[0] *= d;
	s.val[1] *= d;
	s.val[2] *= d;
	s.val[3] *= d;
	s.val[0] += a;
	s.val[1] += a;
	s.val[2] += a;
	s.val[3] += a;
}


//ߐڃsNZƂ̐F̍̍ől臒lr

int checkPoint(IplImage *image, int x, int y, double th)
{
	static struct { int x,y; } map[4] = { { 1,0 }, {-1,0}, { 0,1}, { 0,-1} };
	CvScalar color = getColor(image, x,y);
	for(int i = 0;i<4; i++){
		if (colorLen(color, getColorCheck(image, x+map[i].x, y + map[i].y)) > th) return 1;
	}
	return 0;
}
int setPoint(int index, CvPoint *point, CvScalar &color)
{
	if (index >= MAXPOINT) return 0;
	pointTable[index].p = *point;
	pointTable[index].color.val[0] = color.val[0];
	pointTable[index].color.val[1] = color.val[1];
	pointTable[index].color.val[2] = color.val[2];
	pointTable[index].color.val[3] = color.val[3];
	return 1;		
}
double clip(double x, double min, double max)
{
	if (x>max) return max;
	if (x<min) return min;
	return x;
}
// inputImageAbaseɍ plane 0,1,2:H,S,V 3:GrayScale̎
void drawEdge(IplImage* image,IplImage* base,int plane = 2)
{
	static CvMemStorage *memStorage0 =cvCreateMemStorage(0); 
	CvSeq *contourSeq0 = NULL;

	int height    = image->height;
	int width     = image->width;
	int step      = image->widthStep;
	int channels  = image->nChannels;
	uchar *data      = (uchar *)image->imageData;

	if (plane <3){
		cvCvtColor(image,hsv,CV_BGR2HSV);				// HSVplane搶̌ɂ
		for(int i=0;i<height*width;++i) grayImage->imageData[i] = hsv->imageData[i*3+plane];
	}else{
		cvCvtColor (image, grayImage, CV_BGR2GRAY);		// O[C[W搶̌ɂ
	}

	IplImage *target=base;					// ރ^[QbgC[W


	for(int x=20;x<240;x+=Y){
		cvThreshold( grayImage, binaryImage, x, 255, CV_THRESH_BINARY );	// x ̒lɂQl֊s𒊏o
		contourSeq0 = 0;
		cvFindContours(binaryImage, memStorage0, &contourSeq0,sizeof(CvContour), CV_RETR_LIST, CV_CHAIN_APPROX_NONE, cvPoint(0,0)); // ֊sT
		
		if (lineNoise>0){ // sAC̏ꍇ
			for(; contourSeq0 != 0; contourSeq0 = contourSeq0->h_next) 
			{
				CvPoint *p;
				if (contourSeq0->total<X*5) continue;		// 5p`ȉׂ͔̍̂r
				int index = 0;
				for( int i=0; i<contourSeq0->total; i+=X )
				{
					p = CV_GET_SEQ_ELEM(CvPoint, contourSeq0, i);				// _̏ꏊƐFo^
					CvScalar color = getColor(image,p->x, p->y);
					muladdScaler(color,DARK,addDARK);										// PxC
					color.val[3] = checkPoint(image, p->x, p->y, lineNoise);	// L_ǂߐڃsNZ画f[3]֊i[
					setPoint(index,p,color);									// pointTable֕ۑ
					index++;
					if (index > MAXPOINT){
						printf("INDEX ERROR\n"); 
						index = 0;
					}
				}
				// TAȉ̗L_͖ iPending:j
				for(int i=0;i<index;i++){
					int p1 = i;
					int p2,p3,p4,p0;
					if (pointTable[p1].color.val[3]){
						p2 = (p1+1)%index;
						p3 = (p1+2)%index;
						p4 = (p1+3)%index;
						p0 = (p1-1+index)%index;
						if (pointTable[p0].color.val[3]) continue;
						if (!pointTable[p2].color.val[3] ||
							!pointTable[p3].color.val[3] ||
							!pointTable[p4].color.val[3] ){						
							pointTable[p1].color.val[3] = 0;
						}
					}
				}
				// ڑꂽL_`
				for(int i=0;i<index;i++){
					int p1 = i;
					int p2 = (i+1)%index;// if (p2==index) p2 = 0;
					if (pointTable[p1].color.val[3] && pointTable[p2].color.val[3])
					{
						CvScalar c = pointTable[p1].color;
						muladdScaler(c,DARK,addDARK);
						cvLine(target,pointTable[p1].p,pointTable[p2].p,c,lineWidth,CV_AA );
					}
				}
			}
		}else{
			// S`ꍇ
			for(; contourSeq0 != 0; contourSeq0 = contourSeq0->h_next) 
			{

				CvPoint *p1 = 0;
				CvPoint *p2;

				if (contourSeq0->total<X*5) continue;		

				for(int i=0; i<contourSeq0->total; i+=X ){
					p1 = CV_GET_SEQ_ELEM(CvPoint, contourSeq0, (i)%contourSeq0->total);//n_
					p2 = CV_GET_SEQ_ELEM(CvPoint, contourSeq0, (i+X+Z)%contourSeq0->total);// I_
					CvScalar color = getColor(image,p1->x, p1->y);
					mulScaler(color,DARK);
					cvLine(target,*p1,*p2,color,lineWidth,CV_AA );
				}
			}
		}
	}
	cvClearMemStorage(memStorage0);
}

// HatchingGtFNg
int drawHatching(IplImage *frame, IplImage *base)
{
	cvCvtColor(frame,hsv,CV_BGR2HSV);

	int rx = (int)((double)rand() / ((double)RAND_MAX + 1) * xRange);// texture offset
	int ry = (int)((double)rand() / ((double)RAND_MAX + 1) * yRange);
	double d =  (((double)rand() / ((double)RAND_MAX + 1) - 0.5)*incRange + inclination);//X

	int height    = frame->height;
	int width     = frame->width;

	double a00 = hatchingTf.a00;
	double a01 = hatchingTf.a01;
	double a02 = hatchingTf.a02;
	double a10 = hatchingTf.a10;
	double a11 = hatchingTf.a11;
	double a12 = hatchingTf.a12;

	int hWidth[nPattern];
	int hHeight[nPattern];
	int hWidthStep[nPattern];

	for (int i=0; i<nPattern; i++) {
		hWidth[i] = hatching[i]->width;
		hHeight[i] = hatching[i]->height;
		hWidthStep[i] = hatching[i]->widthStep;
	}

	int vtbl[256];
	for (int i=0; i<256; i++) {
		double V = i/256.0;
		V = pow(V, (10-VFlag)/10.0)*256.0;
		vtbl[i] = (int)V;
	}

	int fnChannels = frame->nChannels;
	int fwidthStep = frame->widthStep;

	for(int y=0;y<height;++y){
		int index_y = y * fwidthStep;

		for(int x=0;x<width;++x){
			int index = index_y + x*fnChannels;
			unsigned char v = hsv->imageData[index+2];
			v = vtbl[v];

			double X = x + rx;
			double Y = y + ry + (x*d);

			double typeV;

			int maxV = nPattern-1+patternOffset;
			
			typeV = (255-v) / (256/maxV)-patternOffset;
			typeV = maxV-typeV ;

			if (typeV<0) typeV = 0;
			if (typeV>nPattern-1) typeV = nPattern-1;

			double vf = 0.0;
			int a = (int)typeV;
			int b = a+1;
			double w = typeV-a; // weight

			if (hatchingType >= 1){ // Inverse
				a = nPattern-1-a;
				b = nPattern-1-b;
			}

			double dx = a00 * X + a01 * Y + a02;
			double dy = a10 * X + a11 * Y + a12;

			switch (hatchingIntrType) {
				case INTR_NONE: {
					if (dx < 0.0) dx--;
					if (dy < 0.0) dy--;

					int aX = modulo((int)(dx+0.5), hatching[a]->width);
					int aY = modulo((int)(dy+0.5), hatching[a]->height);
					vf += ((unsigned char)hatching[a]->imageData[aX + aY * hatching[a]->widthStep]) * (1.0-w);
					if (b >=0 && b<=6){
						int bX = modulo((int)(dx+0.5), hatching[b]->width);
						int bY = modulo((int)(dy+0.5), hatching[b]->height);
						vf += ((unsigned char)hatching[b]->imageData[bX + bY * hatching[b]->widthStep]) * w;

					}
					break;
				}

				case INTR_BILINEAR:
					vf += v_bilinear(hatching[a], dx, dy)  * (1.0-w);
					if (b >=0 && b<=6){
						vf += v_bilinear(hatching[b], dx, dy) * w;
					}
					break;

				case INTR_BICUBIC:
					vf += v_bicubic(hatching[a], dx, dy)  * (1.0-w);
					if (b >=0 && b<=6){
						vf += v_bicubic(hatching[b], dx, dy) * w;
					}
					break;
			}

			if (hatchingType <2)
				vf = 255-vf;

			unsigned char vv = hsv->imageData[index+2];
			if (hatchingType ==2){
				//int V = vv;V*=2;if (V>255)V=255;vv = V;// x[X𖾂邭Ă݂
				vv = (unsigned char)clip(vv*2.0,0.0,255.0);
				vf=(vf*0.7+255.0*0.3); // ̂łƖ邭
			}
			double sf = 1.0;
			if (hatchingType ==3){ // 
				vv = 255;
				sf = 0.0;
				//vf=vf*0.7+255.0*0.3; // ̂łƖ邭
			}
			
			if (hatchingType < 2) {
				vf = 255.0 - clip(255.0-vf,(1-hDenseMax)*255.0, 255.0);
			} else {
				vf = clip(vf,(1-hDenseMax)*255.0, 255.0);
			}

			base->imageData[index+0] = hsv->imageData[index+0];
			base->imageData[index+1] = (unsigned char)(hsv->imageData[index+1]*sf);
			base->imageData[index+2] = (unsigned char)(vv * vf / 255) ;

		}
	}
	cvCvtColor(base,base,CV_HSV2BGR);
	return 1;
}

#if 0
int _tmain( int argc, char** argv )
{
	int id;
	printf("Input device ID = ");scanf("%d",&id);
	CvCapture*	capture = cvCaptureFromCAM( id );
/*
	cvSetCaptureProperty (capture, CV_CAP_PROP_FRAME_WIDTH, 640);
	cvSetCaptureProperty (capture, CV_CAP_PROP_FRAME_HEIGHT, 480);
*/
	if (!capture){
		printf("device open error\n");
		exit(0);
	}

	printf("1/2/3/4... \ސ؂ւ\n");
	printf("5/6    ... hatchingFuncύX\n");
	printf("7/8/9  ... hatchinĝ/Linê/\n");
	printf("f/F    ... FPS\\n");
	printf("ESC    ... I\n");

	IplImage *frame = cvQueryFrame( capture );

	IplImage *res = cvCreateImage (cvGetSize (frame), IPL_DEPTH_8U, 3);
	int width, height;


	width = cvGetSize(frame).width;
	height = cvGetSize(frame).height;
	
	CvFont dfont;
	cvInitFont (&dfont, CV_FONT_HERSHEY_PLAIN  , 1.0f, 1.0f, 0.0f, 1, CV_AA);
	cvNamedWindow( "Hatching", 1 );

	// CLOSE BUTTON@IESC
	HWND hWnd = FindWindow(0,"Hatching");
	HMENU pSysMenu = GetSystemMenu(hWnd,FALSE);
	EnableMenuItem(pSysMenu,SC_CLOSE, MF_GRAYED | MF_BYCOMMAND);


	for(int i=0;i<nPattern;i++){
		char buf[256];
		sprintf(buf,"img/h%d.bmp",i);
		hatching[i] = cvLoadImage(buf);
	}
	DWORD start = timeGetTime();
	int	frameCount = 0;
	char fpsString[64] = " ";
	for(;;)
	{
		int c;

		frame = cvQueryFrame( capture );
		if( !frame ) break;

		if (drawMode & 0x01){
			drawHatching(frame,res);
		}else{
			if (hatchingType>=2){
				cvSet(res,cvScalarAll(255),0);
			}else{
				cvSetZero(res);
			}
		}
		if (drawMode & 0x02){
			drawEdge(frame,res,2);
		}

		// FPS \
		if (drawFps){
			frameCount++;
			DWORD now = timeGetTime();
			if (now-start > 1000){
				float fps = 30000.0f / (now-start);
				sprintf(fpsString,"FPS = %f",fps);
				start = now;
				frameCount = 0;
			}
			cvPutText(res, fpsString, cvPoint(0, height-10), &dfont, CV_RGB(255, 255, 255));
		}

		cvShowImage( "Hatching", res );
		c = cvWaitKey(1);
		if (c==033) break;
		switch(c){
			case '1': hatchingType = 0;DARK=1.0;patternOffset=0;break;
			case '2': hatchingType = 1;DARK=1.0;patternOffset=0;break;
			case '3': hatchingType = 2;DARK=0.7;patternOffset=1;break; // ʔ]Ł@xŗ
			case '4': hatchingType = 3;DARK=0.0;patternOffset=1;break; // 

			case '5': VFlag = 1;break;
			case '6': VFlag = 0;break;
			case '7': drawMode = 0x01;break; // hatching only
			case '8': drawMode = 0x02;break; // line only
			case '9': drawMode = 0x03;break; // both
			case 'f': drawFps  = 1;frameCount = 0;break;
			case 'F': drawFps  = 0;break;
		}
	}
	cvReleaseCapture(&capture);
}

#endif


static
void CopyImgWithMargin(
		IplImage *src, IplImage *dst, int margin
){
	int srcw = src->width;
	int srch = src->height;
	int srcws = src->widthStep;

	int dstw = dst->width;
	int dsth = dst->height;
	int dstws = dst->widthStep;

	/* byte per pixel */
	int bpp = (src->depth / 8) * src->nChannels;

	for (int dsty = 0; dsty < dsth; dsty++) {
		int srcy = modulo(dsty - margin, srch);
		char *psrcy = src->imageData + (srcws * srcy);
		char *pdsty = dst->imageData + (dstws * dsty);

		for (int dstx = 0; dstx < dstw;) {
			if (margin <= dstx && dstx < srcw + margin) {
				char *psrc = psrcy;
				char *pdst = pdsty + margin * bpp;
				memcpy(pdst, psrc, bpp*srcw);
				dstx += srcw;
			} else {
				int srcx = modulo(dstx - margin, srcw);
				char *psrc = psrcy + srcx * bpp;
				char *pdst = pdsty + dstx * bpp;
				memcpy(pdst, psrc, bpp);
				dstx++;
			}
		}
	}
}

static
void AffineTransform(
		IplImage *src, IplImage *dst, CvMat *mat
){
	int srcw = src->width;
	int srch = src->height;
	int srcws = src->widthStep;

	int dstw = dst->width;
	int dsth = dst->height;
	int dstws = dst->widthStep;

	/* byte per pixel */
	int bpp = (src->depth / 8) * src->nChannels;

	double a00 = cvGetReal2D(mat, 0, 0);
	double a01 = cvGetReal2D(mat, 0, 1);
	double a02 = cvGetReal2D(mat, 0, 2);
	double a10 = cvGetReal2D(mat, 1, 0);
	double a11 = cvGetReal2D(mat, 1, 1);
	double a12 = cvGetReal2D(mat, 1, 2);

	for (int dsty = 0; dsty < dsth; dsty++) {
		char *pdsty = dst->imageData + (dstws * dsty);

		for (int dstx = 0; dstx < dstw; dstx++) {
			double x = a00 * dstx + a01 * dsty + a02;
			double y = a10 * dstx + a11 * dsty + a12;

			int xx = modulo((int)x, srcw);
			int yy = modulo((int)y, srch);

			char *psrc = src->imageData + (srcws * yy + bpp * xx);
			char *pdst = pdsty + dstx * bpp;
			memcpy(pdst, psrc, bpp);
		}
	}
}



static char *g_dllpath = NULL;
static std::map<std::string, int> g_trackMap;
static std::map<std::string, int> g_checkMap;

static void ReleaseImages() {
	cvReleaseImage(&img0);
	cvReleaseImage(&img1);
	cvReleaseImage(&hsv);
	cvReleaseImage(&grayImage);
	cvReleaseImage(&binaryImage);
}

static void CreateImages(CvSize size) {
	img0		= cvCreateImage(size, IPL_DEPTH_8U, 3);
	img1		= cvCreateImage(size, IPL_DEPTH_8U, 3);
	hsv			= cvCreateImage(size, IPL_DEPTH_8U, 3);
	grayImage	= cvCreateImage(size, IPL_DEPTH_8U, 1);
	binaryImage	= cvCreateImage(size, IPL_DEPTH_8U, 1);
}

static void ReleaseHatchingPattern() {
	for(int i=0;i<nPattern;i++){
		if (hatching_org[i] != NULL) {
			cvReleaseImage(hatching_org + i);
			hatching_org[i] = NULL;
		}
		if (hatching[i] != NULL) {
			cvReleaseImage(hatching + i);
			hatching[i] = NULL;
		}
	}
}

static bool LoadHatchingPattern(char *errmsg) {
	for(int i=0;i<nPattern;i++){
		char buf[8192];
		if (hatchingPtNo == 0) {
			sprintf(buf,"%s/img/h%d.bmp", g_dllpath, i);
		} else {
			sprintf(buf,"%s/img%d/h%d.bmp", g_dllpath, hatchingPtNo,  i);
		}

		if (hatching_org[i] != NULL) {
			cvReleaseImage(hatching_org + i);
		}

		hatching_org[i] = cvLoadImage(buf, CV_LOAD_IMAGE_GRAYSCALE);

		if (hatching_org[i] == NULL) {
			ReleaseHatchingPattern();
			sprintf(errmsg, "%s ̓ǂݍ݂Ɏs܂", buf);
			return false;
		}
	}
	return true;
}

static void InitHatchingPattern() {
	CvMat *rot = cvCreateMat(3, 3, CV_64FC1);
	CvMat *exp = cvCreateMat(3, 3, CV_64FC1);
	CvMat *cmp = cvCreateMat(3, 3, CV_64FC1);
	CvMat *inv = cvCreateMat(3, 3, CV_64FC1);

	cvSetReal2D(rot, 0, 0, cos(hatchingRad));
	cvSetReal2D(rot, 0, 1, -sin(hatchingRad));
	cvSetReal2D(rot, 0, 2, 0.0);
	cvSetReal2D(rot, 1, 0, sin(hatchingRad));
	cvSetReal2D(rot, 1, 1, cos(hatchingRad));
	cvSetReal2D(rot, 1, 2, 0.0);
	cvSetReal2D(rot, 2, 0, 0.0);
	cvSetReal2D(rot, 2, 1, 0.0);
	cvSetReal2D(rot, 2, 2, 1.0);

	cvSetReal2D(exp, 0, 0, hatchingMagn >= 1.0 ? hatchingMagn : 1.0);
	cvSetReal2D(exp, 0, 1, 0.0);
	cvSetReal2D(exp, 0, 2, 0.0);
	cvSetReal2D(exp, 1, 0, 0.0);
	cvSetReal2D(exp, 1, 1, hatchingMagn >= 1.0 ? hatchingMagn : 1.0);
	cvSetReal2D(exp, 1, 2, 0.0);
	cvSetReal2D(exp, 2, 0, 0.0);
	cvSetReal2D(exp, 2, 1, 0.0);
	cvSetReal2D(exp, 2, 2, 1.0);

	cvMatMul(exp, rot, cmp);
	cvInvert(cmp, inv, CV_LU);

	hatchingTf.a00 = cvGetReal2D(inv, 0, 0);
	hatchingTf.a01 = cvGetReal2D(inv, 0, 1);
	hatchingTf.a02 = cvGetReal2D(inv, 0, 2);
	hatchingTf.a10 = cvGetReal2D(inv, 1, 0);
	hatchingTf.a11 = cvGetReal2D(inv, 1, 1);
	hatchingTf.a12 = cvGetReal2D(inv, 1, 2);

	cvReleaseMat(&rot);
	cvReleaseMat(&exp);
	cvReleaseMat(&cmp);
	cvReleaseMat(&inv);

	for(int i=0;i<nPattern;i++){
		if (hatching[i] != NULL) {
			cvReleaseImage(hatching + i);
		}

		if (hatchingMagn >= 1.0) {
			CvSize size = cvSize(hatching_org[i]->width, hatching_org[i]->height);
			hatching[i] = cvCreateImage(size, hatching_org[i]->depth, hatching_org[i]->nChannels);
			cvCopy(hatching_org[i], hatching[i]);

		} else {	
			CvSize size = cvSize(Saturate((int)(hatching_org[i]->width * hatchingMagn), 1, 2048), 
								 Saturate((int)(hatching_org[i]->height * hatchingMagn), 1, 2048));
			hatching[i] = cvCreateImage(size, hatching_org[i]->depth, hatching_org[i]->nChannels);
			cvResize(hatching_org[i], hatching[i], CV_INTER_AREA);
		}

		continue;
	}
}


void InitAnimEffect(char *dllpath, FILTER_DLL *fdp) {
	g_dllpath = dllpath;

	g_trackMap.clear();
	for (int i=0; i<fdp->track_n; i++) {
		g_trackMap[fdp->track_name[i]] = i;
	}

	g_checkMap.clear();
	for (int i=0; i<fdp->check_n; i++) {
		g_checkMap[fdp->check_name[i]] = i;
	}

	hatchingPtNo = -100;
}

void UpdateAnimEffect(FILTER *fp) {
	int no = fp->track[g_trackMap["݇"]];
	boolean no_changed = false;
	double magn = (double)fp->track[g_trackMap["g嗦"]] / 100.0;
	double rad  = ((double)fp->track[g_trackMap["]px"]]) * M_PI / 180.0;

	if (no != hatchingPtNo) {
		hatchingPtNo = no;
		no_changed = true;

		char errmsg[8192];
		bool ret;
		ret = LoadHatchingPattern(errmsg);
		if (!ret) {
			MessageBox(fp->hwnd, errmsg, "G[", MB_OK);
			return;
		}
	}

	if (hatching_org[0] == NULL) return;
	
	if (no_changed || 
		magn != hatchingMagn ||
		rad != hatchingRad
		) {

		hatchingMagn = magn;
		hatchingRad = rad;

		InitHatchingPattern();
	}

}

void ExitAnimEffect() {
	ReleaseImages();
	ReleaseHatchingPattern();
}


BOOL ExecAnimEffect(FILTER *fp,FILTER_PROC_INFO *fpip) {

	PIXEL_YC *buffer0 = fpip->ycp_edit;
	PIXEL_YC *buffer1 = fpip->ycp_temp;

	int mw = fpip->max_w;
	int mh = fpip->max_h;
	int w = fpip->w;
	int h = fpip->h;

	int hatchingMode = fp->track[g_trackMap["Mode"]];
	param::X = fp->track[g_trackMap["LineStep"]];
	param::Y = fp->track[g_trackMap["oԊu"]];
	param::Z = fp->track[g_trackMap["Length"]];
	lineWidth = fp->track[g_trackMap["Width"]];
	lineNoise = fp->track[g_trackMap["Noise"]];
	VFlag = fp->track[g_trackMap["VFlag"]];
	mDARK = ((double)fp->track[g_trackMap["Zx"]])/100.0;
	inclination = ((double)fp->track[g_trackMap["X"]])/100.0;
	hDenseMax = (fp->track[g_trackMap["Zx"]])/100.0;
	int freq = fp->track[g_trackMap["XVԊu"]];
	int rseed = fp->track[g_trackMap[""]];
	incRange = ((double)fp->track[g_trackMap["Xϓ"]])/100.0;
	xRange = fp->track[g_trackMap["Xϓ"]] + 1;
	yRange = fp->track[g_trackMap["Yϓ"]] + 1;

	int lineOn		= fp->check[g_checkMap["C"]];
	int hatchingOn	= fp->check[g_checkMap["nb`O"]];
	hatchingIntrType = INTR_NONE;
	if (fp->check[g_checkMap["nb`OԂ(޲Ʊ)"]] == 1) {
		hatchingIntrType = INTR_BILINEAR;
	}
	if (fp->check[g_checkMap["nb`OԂ(޲ޯ)"]] == 1) {
		hatchingIntrType = INTR_BICUBIC;
	}

	UpdateAnimEffect(fp);

	switch(hatchingMode){
		case 1: hatchingType = 0;DARK=1.0;patternOffset=0;break;
		case 2: hatchingType = 1;DARK=1.0;patternOffset=0;break;
		case 3: hatchingType = 2;DARK=0.7;patternOffset=1;break; // ʔ]Ł@xŗ
		case 4: hatchingType = 3;DARK=0.0;patternOffset=1;break; // 
	}

	if (hatchingMode != 4) {
		DARK *= mDARK;
		addDARK = 0;
	} else {
		addDARK = 255.0*(1-mDARK);
	}

	if (hatching[0] == NULL) {
		//InitHatchingPattern();
		return FALSE;
	}


	if (img0 == NULL) {
		ReleaseImages();
		CreateImages(cvSize(w,h));
	} else {
		CvSize size = cvGetSize(img0);
		if (size.width != w || size.height != h) {
			ReleaseImages();
			CreateImages(cvSize(w,h));
		}
	}

	CopyBufToCvBGRImg<BYTE>(buffer0, img0, 0,0,w,h, 0,0, mw,mh);

	if (hatchingOn){
		if (freq == 0) {
			srand(rseed);
		} else {
			srand(rseed + fpip->frame / freq);
		}
		rand();
		drawHatching(img0, img1);
	}else{
		if (hatchingType>=2){
			cvSet(img1, cvScalarAll(255),0);
		}else{
			cvSetZero(img1);
		}
	}
	if (lineOn){
		drawEdge(img0,img1,2);
	}

	CopyCvBGRImgToBuf<BYTE>(img1, buffer0, 0,0,w,h, 0,0, mw,mh);

	return TRUE;
}
